package gps_monitor.gps;

import gps_monitor.util.*;
import gps_monitor.gui.*;

/**
 * Import the packages and classes needed by this class.
 * Including javax.comm.* i.e. comm.jar, available from
 * http:\\java.sun.com\ as a standard java extension
 */

import java.io.*;
import java.util.*;
import java.awt.*;
import javax.comm.*;
import java.math.BigDecimal;
/**
 * This class is designed to read input from a GPS receiver.
 * The class will handle
 * GPS sentences input to the serial port, where the data
 * input complies with the NMEA 0183 interface standard.
 * 	This Class currently implements the Runnable, and
 * SerialPortEventListener interfaces. It is largely based upon
 * the SimpleReader class produced by Sun.
 *
 * This class has been developed for Microsoft Windows.  Despite
 * the platform independance of java, input via ports is not platform
 * independant. To use this class you need the javax.comm package.
 *
 * @see javax.comm
 * @author Richard Caudle
 * @version 1.0.4
 */

public class NmeaSerialReader implements Runnable, SerialPortEventListener
{
	static CommPortIdentifier portId;
	static Enumeration portList;
	GpsMonitorUI ui;



	/**
    * Development version number
    */
	public String VERSION_NUMBER = "1.0.4";

	InputStream inputStream;
	InputStreamReader inputSR;
	SerialPort serialPort;
	Thread readerThread;
	String readSentence = "";
	NmeaSerialGps parent;

	/**
	 *  This constructor attempts to create an NmeaSerialReader which will read from
	 * 'inputPort' and whose parent will be a GPS component/widget.
	 *
	 * @see NmeaSerialGps
	 */
	public NmeaSerialReader(String inputPort, int baud, NmeaSerialGps widget, GpsMonitorUI ui) throws PortInUseException, IOException, TooManyListenersException, UnsupportedCommOperationException{

		this.ui = ui;

		try{

			//Set the widget to be this reader's parent
			parent = widget;

			//Retrieve the system's list of ports
			portList = CommPortIdentifier.getPortIdentifiers();

			/*
			 * Here we traverse the list of ports on the system until we
			 * find that which the constructor's parameter requested.
			 * portId ends up being set to this port
			 */
			while (portList.hasMoreElements())
			{
        	   	portId = (CommPortIdentifier) portList.nextElement();

        	   	System.out.println( " ports baby    "+portId.getName() );

				if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL)
				{
					if (portId.getName().equals(inputPort))
					{
						//Port found so quit loop
						break;
        	       	}
        	   	}
        	}
			/* Attempt to open a port, which will be owned by this application.
			 * The second argument is not too important. It just represents
			 * 'timeout', i.e. the time in milliseconds to block waiting for
			 * port to open. This could throw a PortInUseException.
			 */

			 System.out.println("Owner:  "+ portId.getCurrentOwner());
			serialPort = (SerialPort) portId.open("NmeaSerialReader", 2000);



			/* Now attempt to create an InputStream to gather input from our
			 * open port. This could throw an IOException
			 */
			inputStream = serialPort.getInputStream();

			/* Now attempt to add an EventListener to the port, to alert us that
			 * new input has arrived. Could throw a TooManyListenersException.
			 */
			serialPort.addEventListener(this);

			//Set the port to notify us when new data has arrived.
			serialPort.notifyOnDataAvailable(true);

			/* Attempt to set the port's parameters.  The parameters used here
			 * are standard for using the NMEA interface, however they may need to
			 * be changed for different recievers.  In particular some recievers
			 * will have a baud rate of 9600, not 4800. This could throw an
			 * UnsupportedCommOperationException.
			 */
			serialPort.setSerialPortParams(
				4800, 					//Baud rate
				SerialPort.DATABITS_8,	//No. of data bits
				SerialPort.STOPBITS_1,	//No. of stop bits
				SerialPort.PARITY_NONE	//Parity (set to none)
			);

			//Now pipe received data through an InputStreamReader to encode chars from bytes
			inputSR = new InputStreamReader(inputStream);

		}catch(NullPointerException e){e.printStackTrace();}

		/* Create a new thread which will encapsulate the object we are
		 * creating. Once this is done we start the thread running.
		 */
        readerThread = new Thread(this);
        readerThread.start();
    }



	/**
	 * This method implements the run() method needed by runnable.  The thread
	 * we are using is set to sleep for 20 seconds, but really until it is
	 * interrupted by the arrival of the first nmea data on the port.
	 *
	 * @see java.lang.Runnable
	 */
    public void run()
	{
        try
		{
            Thread.sleep(20000);
        }
		catch (InterruptedException e)
		{}
    }

	/**
	 * This method implements the serialEvent() method required by the
	 * SerialPortEventListener interface.  It describes what action should be taken
	 * when an event occurs on the port.
	 *
	 * @see javax.comm.SerialPortEventListener
	 * @param event The name of the event that has occured on the serial port.
	 */
    public void serialEvent(SerialPortEvent event)
	{



		/* The switch structure here is only really interested in one of
		 * the possible events that can occur on the port. We want to know
		 * when new data arrives (DATA_AVAILABLE).  If any of the others
		 * occur then the statement is broken, i.e. the event is ignored.
		 */
        switch(event.getEventType())
		{
        	case SerialPortEvent.BI:
        	case SerialPortEvent.OE:
        	case SerialPortEvent.FE:
	        case SerialPortEvent.PE:
    	    case SerialPortEvent.CD:
        	case SerialPortEvent.CTS:
	        case SerialPortEvent.DSR:
    	    case SerialPortEvent.RI:
        	case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
            	break;

			//Defines what should happen when new data is recieved.
        	case SerialPortEvent.DATA_AVAILABLE:
            {
				char[] readChar = {'*'};
				char currentChar;
				int intIgnored = 0;

				/* Exceptions can occur whilst parsing Strings (NullPointerExcpetions),
				 * and gathering data from the InputStreamReader (IOExceptions).
				 * However these can not be thrown by the method otherwise we will not
				 * be correctly implementeing serialEvent in SerialPortEvent.
				 */
				 try
				 {
				 	//Do this while the input stream is not empty
                	while (inputSR.ready())
					{
						//Read the first character from the stream
						intIgnored = inputSR.read(readChar, 0, 1);
						currentChar = readChar[0];

						//Loop until a '$' symbol is found, indicating the start of a sentence
						while (currentChar != '$')
						{
							intIgnored = inputSR.read(readChar, 0, 1);
							currentChar = readChar[0];
						}

						//Put $ bacl on to the start of our sentence
						readSentence = "$";

						//Do this forever (actually until return char found)
						while (true)
						{
							//Get next char
							intIgnored = inputSR.read(readChar, 0, 1);
							currentChar = readChar[0];

							//If return char decode sentence and exit this loop
							if ((currentChar == (char)13))
							{
								//System.out.println(readSentence);
								ui.setStat(readSentence);

								decodeNmea(readSentence);
								//Skip line feed char
								intIgnored = inputSR.read();
								break;
							}
							//Otherwise add the char to the current sentence
							else
							{
								readSentence += currentChar;
							}
						}
                	}
				}
				catch (IOException e)
				{}
				catch (NullPointerException e)
				{}
				//Exit switch statement
				break;
	        }
        }
	}



	/**
	 * This method is used to parse sentences that are input to the port being listened
	 * to. The sentences will only be passed successfully if they conform to the
	 * NMEA 0183 v2 protocol.  If a sentence is not recognised it is ignored.
	 *
	 * @param sentence The sentence that should be parsed.
	 */
	private void decodeNmea(String sentence)
	{

		/*
		 * In the following parsing activities these exceptions are likely
		 * to be thrown.
		 *  NullPointerException??
		 *  NumberFormatException - whilst creating numeric variables from Strings
		 *  StringIndexOutOfBoundsException - whilst moving through a String (if it is an invalid String)
		 */
		 try
		 {
		 	//Remove $ from start of sentence
			String inputSentence = sentence.substring(1);
			//Variable to store the index of the next comma
			int nextComma = 0;

			//These sentences contain route data and hence are ignored
			if ((inputSentence.startsWith("GPBOD"))||(inputSentence.startsWith("GPRTE")))
			{
				//System.out.println("Route details have been ignored");
			}

			//To deal with a GPGLL sentence
			else if (inputSentence.startsWith("GPGLL"))
			{

				LatOrLongitude latitude =null, longitude= null;

				inputSentence = inputSentence.substring(6);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get latitude
					if ((inputSentence.charAt(nextComma + 1)) == 'N')
						latitude = new LatOrLongitude(inputSentence.substring(0,nextComma));
					else
						latitude = new LatOrLongitude("-" + inputSentence.substring(0,nextComma));
					//Update parent
					parent.setLatitude(latitude);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');
				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get longitude
					if((inputSentence.charAt(nextComma + 1)) == 'E')
						longitude = new LatOrLongitude(inputSentence.substring(0,nextComma));
					else
						longitude = new LatOrLongitude("-" + inputSentence.substring(0,nextComma));
					//Update parent
					parent.setLongitude(longitude);
				}


			}

			//To deal with a GPRMC sentence
			else if (inputSentence.startsWith("GPRMC"))
			{
				inputSentence = inputSentence.substring(6);
				nextComma = inputSentence.indexOf(',');

				LatOrLongitude latitude=null, longitude=null;
				double speed, heading;
				boolean gotTime = false;
				int hours = -1;
				int mins = -1;
				int secs = -1;

				if (nextComma != 0)
				{
					//Get time data received
					String time = inputSentence.substring(0,nextComma);
/*  some receivers put fractions of a second on afterwards.  we'll just discard them.
					if (time.length() != 6)
						throw new NumberFormatException("Invalid time received.");
*/
					hours = (new Integer(time.substring(0,2))).intValue();
					mins = (new Integer(time.substring(2,4))).intValue();
					secs = (new Integer(time.substring(4,6))).intValue();
					gotTime = true;
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get status of receiver (active, or out of range only)
					char statusChar = inputSentence.charAt(0);
					//Send data to parent
					switch (statusChar)
					{
						case 'A':	parent.setStatus(NmeaSerialGps.ACTIVE);
						case 'V':   parent.setStatus(NmeaSerialGps.RECEPTION_WARNING);
						default:
					}
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get latitude
					if ((inputSentence.charAt(nextComma + 1)) == 'N')
						latitude = new LatOrLongitude(inputSentence.substring(0,nextComma));
					else
						latitude = new LatOrLongitude("-" + inputSentence.substring(0,nextComma));
					//Update parent
					parent.setLatitude(latitude);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');
				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//get longitude
					if((inputSentence.charAt(nextComma + 1)) == 'E')
						longitude = new LatOrLongitude(inputSentence.substring(0,nextComma));
					else
						longitude = new LatOrLongitude("-" + inputSentence.substring(0,nextComma));
					//Update parent
					parent.setLongitude(longitude);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');
				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get speed
					String speedString = inputSentence.substring(0,nextComma);
					//convert speed to the required metric unit (km/hour)
					BigDecimal knotsToMetres = new BigDecimal("1.852");
					speed = ((new BigDecimal(speedString)).multiply(knotsToMetres)).doubleValue();
					//Update parent
					parent.setSpeed(speed);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//System.out.println("heading " + inputSentence.substring(0,nextComma));
					//get heading (note this is the true heading not magnetic)
					heading = (new Double(inputSentence.substring(0,nextComma))).doubleValue();
					//Update parent
					parent.setTrueHeading(heading);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get the date
					String dateNow = inputSentence.substring(0, nextComma);
					//System.out.println("date " + dateNow);
					if (dateNow.length() != 6)
						throw new NumberFormatException("Invalid date received.");

					int date = (new Integer(dateNow.substring(0,2))).intValue();
					int month = (new Integer(dateNow.substring(2,4))).intValue();
					int year = (new Integer(dateNow.substring(4,6))).intValue();

					if (gotTime)
						parent.setTimeLastLocationUpdate(year, month, date, hours, mins, secs);
				}



			}

			//To deal with a GPRMB sentence
			else if(inputSentence.startsWith("GPRMB"))
			{}

			//To deal with a GPGGA sentence
			else if(inputSentence.startsWith("GPGGA"))
			{
				inputSentence = inputSentence.substring(6);
				nextComma = inputSentence.indexOf(',');

				LatOrLongitude latitude=null, longitude=null;
				long timeReceived;


				 //  Removed because we have no way of getting the date from this sentence

				 //Get time data received


				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get latitude
					if ((inputSentence.charAt(nextComma + 1)) == 'N')
						latitude = new LatOrLongitude(inputSentence.substring(0,nextComma));
					else
						latitude = new LatOrLongitude("-" + inputSentence.substring(0,nextComma));
					//Update receiver
					parent.setLatitude(latitude);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');
				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//get longitude
					if((inputSentence.charAt(nextComma + 1)) == 'E')
						longitude = new LatOrLongitude(inputSentence.substring(0,nextComma));
					else
						longitude = new LatOrLongitude("-" + inputSentence.substring(0,nextComma));
					//Update parent
					parent.setLongitude(longitude);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');
				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Get the number of satellites used
					int satellitesUsed = (new Integer(inputSentence.substring(0,nextComma))).intValue();
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');

				if (nextComma != 0)
				{
					//Gather HDOP info
					double hdop = (new Double(inputSentence.substring(0,nextComma))).doubleValue();
					//Update parent
					parent.setHorizontalDOP(hdop);
				}

				inputSentence = inputSentence.substring(nextComma + 1);
				nextComma = inputSentence.indexOf(',');
				if (nextComma != 0)
				{
					//Now gather information on altitude of receiver
					double altitude = (new Double(inputSentence.substring(0,nextComma))).doubleValue();
					//Update parent
					parent.setAltitude(altitude);
				}

				//Information follows regarding altitude accuracy

				//try{
				//Pos pos = new Pos();
				//pos.lat = Double.parseDouble(latitude.toStringDDMM());
				//pos.lng = Double.parseDouble(longitude.toStringDDMM());
        		//Point p=d.transPos(pos);
        		//w.updatePos(p,d.transPos(calcPred(pos)));
				//}catch(Exception e){System.out.println(" zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz");}


			}

			//To deal with a GPGSA sentence
			else if(inputSentence.startsWith("GPGSA"))
			{}

			//To deal with a GPGSV sentence
			else if(inputSentence.startsWith("GPGSV"))
			{}

			//To deal with a GPAPB sentence
			else if(inputSentence.startsWith("GPAPB"))
			{}

			//To deal with a PRWIZCH sentence (rockwell receiver)
			else if(inputSentence.startsWith("PRWIZCH"))
			{}


			//If the sentence is not recognised by this reader
			else
			{
				System.out.println("Unrecognised/non-standard sentence: " + sentence);
			}
		}
		catch (NumberFormatException e)
		{
			System.out.println("NumberFormatException due to input: " + e);
		}
		catch (StringIndexOutOfBoundsException e)
		{
			System.out.println("StringIndexOutOfBoundsException: Input sentence was incorrectly formatted.");
		}
	}





}