package gps_monitor.util;

import java.text.DecimalFormat;
import java.math.BigDecimal;

/**
 * This class has been created to help provide an abstract representation
 * of a GPS/global coordinate. It provides a pure Java representation
 * of either a latitude measurement, or a lonigtude measurement. It is not
 * intended to oppose the Java3D concept.
 *
 * I felt it necessary to write this class as opposed to simply use the
 * Java3D package.  This is because I did not want users to have to install
 * the non-standard package for my use of one class, this seemed excessive.
 * Also a geosptial coordinate is not conceptually compatible with the Java3D
 * coordinate, which is of cartesian form.
 *
 * This class assumes that when regarding latitude, 'North' is represented
 * by a positive value, and 'South' is represented by a negative value.  With
 * longitude 'East' is positive, and 'West' is negative.
 *
 * @see locationUtils.GlobalCoordinate
 * @author Richard Caudle
 * @version 1.0.0
 */

public class LatOrLongitude
{

	/**
	 * Declare the variables that will be used to store the values
	 * needed by this class.
	 */
	double degrees;
	double minutes;
	double seconds;

	/**
	 * This contstructor allows us to create an object from a String
	 * that may obtained from a NMEA sentence for example.
	 * It assumes that the format of the data is DDMM.MM.
	 *
	 * @param input A string representation of the measurement
	 */
	public LatOrLongitude(String input) throws NumberFormatException, StringIndexOutOfBoundsException
	{
		int point = input.indexOf('.');

		if (point != -1)
		{
			seconds = (new Double(input.substring(point))).doubleValue() * 60;
			minutes = (new Double(input.substring(point - 2))).doubleValue();
			degrees = (new Integer(input.substring(0,point - 2))).intValue() + minutes/60;
			checkValid();
		}
		else
		{
			seconds = 0.0;
			minutes = (new Double(input.substring(input.length() - 2))).doubleValue();
			degrees = (new Integer(input.substring(0, input.length() - 2))).intValue() + minutes/60;
		}

	}

	/**
	 * This constructor allows us to create an object using three
	 * arguments; one for degrees, one for minutes and the last for
	 * seconds.  The final argument is a double so that we are not
	 * limited to just six figures for our position. This method is
	 * clearly similar to other constructors, but particularly aids
	 * those using the DDMMSS.SS format.
	 *
	 * @param degrees The degrees 'element' of the measurement
	 * @param mintues The minutes 'element' of the measurement
	 * @param seconds The seconds 'element' of the measurement
	 */
	public LatOrLongitude(int degs, int mins, double secs) throws NumberFormatException
	{
		minutes = mins + secs/60;
		seconds = secs;
		degrees = degs + minutes/60;
		checkValid();
	}

	/**
	 * This constructor allows us to create an object using two
	 * arguments; one for degrees, one for minutes.  The final argument
	 * is a double so that we are not limited to just four figures for
	 * our position. This method is similar to other constructors,
	 * but aids those using the DDMM.MM format.
	 *
	 * @param degrees The degrees 'element' of the measurement
	 * @param mintues The minutes 'element' of the measurement
	 */
	public LatOrLongitude(int degs, double mins) throws NumberFormatException
	{
		minutes = mins;
		degrees = degs + minutes/60;

		/* I must admit this is not an elegant solution to obtaining
		 * the seconds value. Any other attempts are welcome!
		 * I have done it in this way to avoid errors in floating point
		 * multiplication that often occurs.
		 */
		String temp = String.valueOf(mins);
		temp = temp.substring(temp.indexOf('.'));
		seconds = ((new BigDecimal(temp)).multiply(new BigDecimal("60"))).doubleValue();

		checkValid();
	}



	/**
	 * This method outputs the measurement in the DDMM.MM format
	 *
	 * @return The measurement of latitude/longitude as a String
	 */
	public String toStringConventional()
	{
		DecimalFormat minFormatter = new DecimalFormat("00.###");
		DecimalFormat degFormatter = new DecimalFormat("#00");
		return new String(degFormatter.format(degrees-1) + "" + minFormatter.format(minutes));
	}

	/**
	 * This method outputs the measurement in the DDMM.MM format
	 *
	 * @return The measurement of latitude/longitude as a String
	 */
	public String toStringDDMM()
	{
		DecimalFormat minFormatter = new DecimalFormat("00.################");
		DecimalFormat degFormatter = new DecimalFormat("#00");
		return new String(degFormatter.format(degrees) + minFormatter.format(minutes));
	}

	/**
	 * This method outputs the measurement in the DDMMSS.SS format
	 *
	 * @return The measurement of latitude/longitude as a String
	 */
	public String toStringDDMMSS()
	{
		DecimalFormat secFormatter = new DecimalFormat("00.################");
		DecimalFormat minFormatter = new DecimalFormat("00");
		DecimalFormat degFormatter = new DecimalFormat("#00");
		return new String(degFormatter.format(degrees) + minFormatter.format(Math.floor(minutes)) + secFormatter.format(seconds));
	}

	/**
	 * This method returns the number of degrees in this LatOrLongitude object.
	 * Note this is not an integer value because it includes the minutes and
	 * seconds values within the returned value.  Casting the value to an int
	 * will give only the number of degrees.
	 *
	 * @return The degrees in this object/measurement
	 */
	public double getDegrees()
	{
		return degrees;
	}

	/**
	 * This method returns the number of minutes in this LatOrLongitude object.
	 * Note as I assume the DDMM.MM format this returns a double, not an int.  If
	 * you are using the DDMMSS.SS format simply cast the result of this method
	 * to an int, this will be the minutes 'element' you require.
	 *
	 * @return The minutes in this object/measurement
	 */
	public double getMinutes()
	{
		return minutes;
	}

	/**
	 * This method returns the number of degrees in this LatOrLongitude object.
	 * It is provided for users of the DDMMSS.SS format.
	 *
	 * @return The seconds in this object/measurement
	 */
	public double getSeconds()
	{
		return seconds;
	}

	/**
	 * A method to test whether two LatOrLong objects are equal.
	 *
	 * @param obj The object which this will compared to.
	 * @return Whether the two objects are equal
	 */
	public boolean equals(LatOrLongitude obj)
	{
		if (obj != null)
		  return (obj.getDegrees() == degrees) &&
			   (obj.getMinutes() == minutes) &&
			   (obj.getSeconds() == seconds);
		else
		  return false;
	}

	/**
	 * A method to test whether one LatOrLong objects is greater than another.
	 * for Latitude, this means more North
	 * for Longitude, this means more East.  Be careful at the international dateline.
	 *
	 * @param obj The object which this will compared to.
	 * @return Whether the two objects are equal
	 */
	public boolean greaterThan(LatOrLongitude obj)
	{
		boolean retval = false;
		if (obj != null){
			if (obj.getDegrees() > degrees){
				retval = true;
			}else
			if ((obj.getDegrees() == degrees) &&
			    (obj.getMinutes() > minutes)){
					retval = true;
			}else
			if ((obj.getDegrees() == degrees) &&
			    (obj.getMinutes() == minutes) &&
			    (obj.getSeconds() > seconds)) {
					retval = true;
			}
		}
		return retval;
	}



	/**
	 * Checks that the number of minutes and seconds are in the valid range
	 * and hence whether the Latitude/Longitude object is valid.  If not it
	 * raises a NumberFormatException if not.
	 */
	private void checkValid() throws NumberFormatException
	{
		if ((!(minutes <= 60.0))||(!(seconds <= 60.0)))
		{
			throw new NumberFormatException("Invalid latitude/longitude data: mins or secs out of bounds");
		}
	}

	/**
	 * Adds an offset to the degree.  Used to correct apparent discrepancy in GpsGenerator generated values
	 *
	*/
	public void offsetDegrees(int offset){
		degrees += offset;
	}

}