//package locationUtils;
/**
 * This class contains algorithms for converting between coordinate/location systems
 *
 * @author Richard Caudle
 * @version 1.0.0
 */

package gps_monitor.util;

public class LocationConverter
{

   /**
	*  These constants are needed by the method which converts the position
	* data to UTM format.  They are currently set to conform with the WGS84
	* ellipsoid which is the global standard.  Consult this sight if you wish
	* to set them to a more appropriate ellipsoid
	*
	*    http://www.colorado.edu/geography/gcraft/notes/datum/edlist.html
	*
	*  Central Scale Factor - k0
	*  Semi-major axis - a
	*  1/flattening - flat
    */
	static double k0 = 0.9996;
	static double flat = 298.257223563;
	static double a = 6378137.0;
	static double degreesToRadians = Math.PI / 180;

    /**
	 *  This method converts the latitude and longitude of the current position to UTM
	 * coordinates. It uses Redfearn's formulae as it's basis.  Note the method
	 * assumes the input coordinates are according to the WGS84 datum, and will
	 * output according to that datum too.
	 *  This website is a good reference regarding the coordinate system:
	 *
	 *     http://www.colorado.edu/geography/gcraft/notes/coordsys/coordsys_f.html
	 */

	public static String geospatialToUTM(LatOrLongitude lat, LatOrLongitude lon)
	{
		double e2 = (2 * (1/flat)) - ((1/flat) * (1/flat));
		double e2i = e2 / (1 - e2);
		double latInRadians = (lat.getDegrees()) * degreesToRadians;
		double longInRadians = (lon.getDegrees()) * degreesToRadians;
		double tempLong = (lon.getDegrees() + 180) - (Math.floor((lon.getDegrees() + 180)/360) * 360) - 180;
		double zoneNo = Math.floor((tempLong + 180)/6) + 1;
		double longOrigin = 6 * (zoneNo - 1) - 180 + 3;
		double originInRadians = longOrigin * degreesToRadians;

		double N = a / Math.sqrt((1-(e2 * Math.sin(latInRadians) * Math.sin(latInRadians))));
		double T = (Math.tan(latInRadians)) * (Math.tan(latInRadians));
		double C = e2i * Math.cos(latInRadians) * Math.cos(latInRadians);
		double A = (longInRadians - originInRadians) * Math.cos(latInRadians);
		double M = a * (
		            ((1 - e2/4 - (3 * e2 * e2)/64 - (5 * e2 * e2 * e2)/256) * latInRadians)
					- (((3 * e2)/8 + (3 * e2 * e2)/32 + (45 * e2 * e2 * e2)/1024) * Math.sin(2 * latInRadians))
					+ (((15 * e2 * e2)/256 + (45 * e2 * e2 * e2)/1024) * Math.sin(4 * latInRadians))
					- (((35 * e2 * e2 * e2)/3072) * Math.sin(6 * latInRadians))
		           );

		double easting = (k0 * N *
					(A
						+ ((1 - T + C) * A * A * A)/6
						+ ((5 - (18 * T) + (T * T) + (72 * C) - (58 *e2i)) * A * A * A * A * A)/120))
						+ 500000;

		double northing = k0 * (
					M + (N * Math.tan(latInRadians) * (
					  (A * A / 2)
					  + ((5 - T + (9 * C) + (4 * C * C)) * A * A * A * A)/24
					  + ((61 - (58 * T) + (T * T) + (600 * C) - (330 * e2i)) * A * A * A * A * A * A)/720
					)));

		if (lat.getDegrees() < 0)
			northing = northing + 10000000;

		return new String(String.valueOf((int)zoneNo) + " " + UTMFindLetter(lat.getDegrees()) + " " + easting + " " + northing);
	}

	/**
	 * This method can be used to convert geospatial coordinates to OS national grid
	 * references suitable for use in the UK. To aquire accurate results the
	 * input coordinates must be according to the Airy1830 and not the WGS84 ellipsoid.
	 *
	 * @param lat The latitude of the 2D coordinate
	 * @param lon The longitude of the 2D coordinate
	 * @return The OS grid reference
	 */
	public static String geospatialToOSGB(LatOrLongitude lat, LatOrLongitude lon)
	{
		double[] coords = geospatialToCOORDS(lat, lon);
		double easting = coords[0];
		double northing = coords[1];

		easting = easting + 1000000;
		northing = northing + 500000;

		int i = (int) Math.floor(easting/500000);
		int j = (int) Math.floor(northing/500000);

		int temp = i + j * 5;
		String L = (new String("VWXYZQRSTULMNOPFGHJKABCDE")).substring(temp, temp + 1);

		easting = easting % 500000;
		northing = northing % 500000;

		i = (int) Math.floor(easting/100000);
		j = (int) Math.floor(northing/100000);

		temp = j * 5 + i;

		L = L + (new String("VWXYZQRSTULMNOPFGHJKABCDE")).substring(temp, temp + 1);

		easting = easting % 100000;
		northing = northing % 100000;

		String eastString = String.valueOf((int) (100000 + easting));
		String northString = String.valueOf((int) (100000 + northing));

		L = L + " " + eastString.substring((eastString.length() - 5), eastString.length()) + " " + northString.substring((northString.length() - 5), eastString.length());

		return L;
	}

	/**
	 * This method is used to convert 2D geospatial coordinates to cartesian
	 * coordinates. The method returns a double array, in which the first element is
	 * the easting(X) value and the second the northing value(Y).
	 *
	 * @param lat The latitude of the 2D coordinate
	 * @param lon The longitude of the 2D coordinate
	 * @return An array containing the cartesian coordinate.
	 */
    public static double[] geospatialToCOORDS(LatOrLongitude lat, LatOrLongitude lon)
    {
		//define a suitable origin
		double longitudeOrigin = -2;
		double latitudeOrigin = 49;
		double eastingOrigin = -400000;
		double northingOrigin = 5527056;
		double majoraxis = 6377563.396;
		double minoraxis = 6356256.9;
		double majsq = majoraxis * majoraxis;
		double minsq = minoraxis * minoraxis;

		//rad# = Atn(1) / 45
		double eccsq = (majsq - minsq) / majsq;
		double ecc4 = eccsq * eccsq;
		double ecc6 = eccsq* ecc4;

		//meridinal distance parameters
		double PA = 1 + 0.75 * eccsq + 0.703125 * ecc4 + 0.068359375 * ecc6;
		double PB = 0.75 * eccsq + 0.9375 * ecc4 + 1.025390625 * ecc6;
		double PC = 0.234375 * ecc4 + 0.41015625 * ecc6;

		double lati = (lat.getDegrees()) * degreesToRadians;
		double longi = degreesToRadians * (lon.getDegrees() - longitudeOrigin);
		double t = Math.tan(lati);
		double tsq = t*t;

		double nu = majoraxis / Math.sqrt((1-(eccsq * Math.sin(lati) * Math.sin(lati))));
		double L = longi * Math.cos(lati);

		double etasq = (Math.cos(lati)) * (Math.cos(lati)) * eccsq;

		//t! = Tan(lati!)

		//compute easting

		double A7 = (61 - tsq * (479 - 179 * tsq + tsq * tsq)) / 5040;
		double A5 = (5 - tsq * (18 - tsq) + etasq * (14 - 58 * tsq)) / 120;
		double A3 = (1 - tsq + etasq) / 6;
		double A1 = k0 * nu;
		double easting = A1 * L * (1 + L * L * (A3 + L * L * (A5 + A7 * L * L)));

		//now do northing
		double m = majoraxis * (1 - eccsq) * (PA * lati - 0.5 * PB * Math.sin(2 * lati) + 0.25 * PC * Math.sin(4 * lati));
		double A2 = 0.5 * A1 * t;
		double A4 = (5 - tsq + etasq * (9 + 4 * etasq)) / 12;
		double A6 = (61 - tsq * (58 - tsq) + etasq * (270 - 330 * tsq)) / 360;
		double northing = k0 * m + A2 * L * L * (1 + L * L * (A4 + A6 * L * L));

		northing = northing - northingOrigin;
		easting = easting - eastingOrigin;

		double[] result = {easting, northing};
		return result;
    }

	/**
	 * This method is used to establish which latitudinal zone the
	 * given northing lies in.
	 */
	private static char UTMFindLetter(double lat)
	{

		char letter;

		double area = (lat + 80)/8;
		//System.out.println(area);

		if ((area < 0) || (area > 164))
		{
			throw new NullPointerException("No valid UTM zone");
		}
		else
			letter = (char) (Math.floor(area) + 2 + 67);

		return letter;
	}
}
